package filter
{
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.Graphics;
	import flash.display.Shader;
	import flash.display.ShaderJob;
	import flash.display.ShaderParameter;
	import flash.utils.getTimer;
	
	public class SplineFilter 
	{
		
		/**	The four points for a Bezier Curve
		 * 	type: any Class with properties .x and .y 
		 * */
		public var
			startpoint:*,
			control1:*,
			control2:*,
			endpoint:*;
		
		
		/** source image */
		protected var 
			_input:BitmapData,
			_divided:Boolean=true,
			_repeat:Boolean=true,
			_fit:uint=3;
			
			
		/**	destination image */
		protected var	_output:BitmapData;
		
		
		/**	embedded .pbj file */
		[Embed(source='shaders/BezierAligner.pbj', mimeType = 'application/octet-stream')]
		protected var EmbeddedShader:Class;
		
		
		/**	shader */
		protected var shader:Shader= new Shader(new EmbeddedShader());
		
		
		/**	For testing */
		public var renderingTime:Number=0.0;
		
		
		
		/**	Constructor */
		public function SplineFilter(
			_input:BitmapData,
			_output:BitmapData)
		{
			input=_input;
			output=_output;
		}
		
		
		/**	Applies the shader to image */
		public function update():void {
			if (_input==null || _output==null) return;
			
			shader.data.background.input = _output;
			shader.data.texture.input = _input;
			
			var shaderJob:ShaderJob = new ShaderJob(shader, _output, _output.width, _output.height);
			
			//	apply 'fit', a number how many time image repeats along the spline:
			if (_fit>0) {
				var sc:Number = this.length/this._input.width/_fit;
				shader.data.scale.value = [sc,sc];
			}
			
			this.renderingTime = -getTimer();
			_output.lock();
			_output.fillRect(_output.rect, 0x0);
			
			//	turns: an array of t-values where path changes its direction:
			var turns:Array = this._divided ? this.turnpoints() : [];
			
			shader.data.tstart.value= [ 0 ];
			if (turns.length==1) {
				//	with one turnpoint:
				//	render first the sequence from 0 to tx
				var tx:Number = turns[0];
				shader.data.tend.value= [ tx ];
				shaderJob.start(true);
				
				//	then the second part from tx to 1
				shaderJob = new ShaderJob(shader, _output, _output.width, _output.height);
				shader.data.tstart.value= [ tx ];
				
			} else if (turns.length==2) {
				//	first part from startpoint to turns[0]
				shader.data.tend.value= [ turns[0] ];
				shaderJob.start(true);
				
				//	second part from turns[0] to turns[1]
				shaderJob = new ShaderJob(shader, _output, _output.width, _output.height);
				shader.data.tstart.value= [ turns[0] ];
				shader.data.tend.value= [ turns[1] ];
				shaderJob.start(true);
				
				//	third part from turns[1] to endpoint
				shaderJob = new ShaderJob(shader, _output, _output.width, _output.height);
				shader.data.tstart.value= [ turns[1] ];
			} 
			
			//	this part renders the last part of path, always to the endpoint t=1.0
			shader.data.tend.value= [ 1 ];
			
			//	apply the shader, true = waits here for completion
			shaderJob.start(true);
			
			_output.unlock();
			this.renderingTime += getTimer();
		}
		
		
		/**	Draws a corresponding Bezier as a line on graphics */
		public function drawPreview(g:Graphics):void {
			var fx:Array =this.bezierEquation("x"),
				fy:Array =this.bezierEquation("y")
				;
			
			g.moveTo(this.control1.x, this.control1.y);
			g.lineTo(fx[0],fy[0]);
			for (var t:Number =0.0; t<1.001; t+= 0.05) {
				g.lineTo(
					fx[0]+t*(fx[1]+t*(fx[2]+t*fx[3])),
					fy[0]+t*(fy[1]+t*(fy[2]+t*fy[3]))
					);
			}
			g.lineTo(this.control2.x, this.control2.y);
		}
		
		
		/**	Reads the four points from the sources and updates the path */
		public function updatePath():void {
			shader.data.startpoint.value= [this.startpoint.x, this.startpoint.y];
			shader.data.control1.value= [this.control1.x, this.control1.y];
			shader.data.control2.value= [this.control2.x, this.control2.y];
			shader.data.endpoint.value= [this.endpoint.x, this.endpoint.y];
			update();
		}
		
		
		/**	Returns an approximation of curve's length along the path,
		 * 	needed for fitting the image's width with the curve */
		public function get length():Number {
			/*
			//	do the same calculations as in pbk-code:
			
			#define splinederivative2(t) float2(eval(dfx,t),eval(dfy,t))
			#define ds(t) length(splinederivative2(t))
			
			tmp = ds(0.0)+
				3.0*(ds(0.33333333*t)+
					ds(0.66666666*t))+
				ds(t);
			
			p2.x = 0.125*t*tmp;
			
			*/
			var fx:Array =this.derivative("x"),
				fy:Array =this.derivative("y"),
				len:Number = 0.0;
			
			//	Using the Simpson method for approximating the length:
			for (var i:int=0; i<4; i++) {
				var t:Number = i/3,
					dx:Number = fx[0] +t*(fx[1] +t*(fx[2] +t*fx[3])),
					dy:Number = fy[0] +t*(fy[1] +t*(fy[2] +t*fy[3]));
				len += (i==0 || i==3 ? 1.0 : 3.0) * // weight: 1 and start and end, 3 in between
					Math.sqrt(dx*dx+dy*dy);
			}
			return 0.125*len; // divide by weights 1+3+3+1
		}
		
		
		/**	Counts the coefficients for the function x(t) or y(t) 
		 * in case of a Quadratic spline 
		 */
		public function quadraticEquation(prop:String="x"):Array { 
			var x0:Number = startpoint[prop],
				controlX1:Number = control1[prop],
				x:Number = endpoint[prop];
			return [ x0, 2*(controlX1-x0), x+x0-2*controlX1 ];
		}
		
		
		/**	Counts the coefficients for the function x(t) or y(t) 
		 * in case of a Bezier spline 
		 */
		public function bezierEquation(prop:String="x"):Array { 
			var x0:Number = startpoint[prop],
				controlX1:Number = control1[prop],
				controlX2:Number = control2[prop],
				x:Number = endpoint[prop];
			return [ x0, 3*(controlX1-x0), 3*(x0-2*controlX1+controlX2), x-x0+3*(controlX1-controlX2)];
		}
		
		
		/**	Counts the coefficients for x'(t) or y'(t) 
		 * in case of a Bezier spline 
		 */
		private function derivative(prop:String="x"):Array { 
			var x0:Number = startpoint[prop],
				controlX1:Number = control1[prop],
				controlX2:Number = control2[prop],
				x:Number = endpoint[prop];
			return [ 3*(controlX1-x0), 6*(x0-2*controlX1+controlX2), 3*(x-x0+3*(controlX1-controlX2)), 0];
		}
		
		
		/**	Set the source bitmapdata */
		public function set input(value:BitmapData):void {
			this._input = value;
			if (value!=null) {
				shader.data.offset.value[1] = value.height/2;
				if(this.repeat) shader.data.imagewidth.value= [ value.width ];
			}
			update();
		}
		public function get input():BitmapData {
			return this._input;
		}
		
		
		/**	Set the destination bitmapdata */
		public function set output(value:BitmapData):void {
			this._output = value;
			update();
		}
		public function get output():BitmapData {
			return this._output;
		}
		
		
		
		public function secondZeros():String {
			var fx:Array = this.bezierEquation("x"),
				fy:Array = this.bezierEquation("y");
			// f(x)= 	A +Bx +Cxx +Dxxx
			// f'(x)= 	B +2Cx +3Dxx
			// f''(x)=	2C +6Dx = 0
			var x0:Number = -fx[2] / (3*fx[3]),
				y0:Number = -fy[2] / (3*fy[3]),
				res:String="";
			
			if (0<x0 && x0<1) res += "x0: "+x0;
			if (0<y0 && y0<1) res += "\ny0: "+y0;
			
			return res;
		}
		
		
		/**	To correctly get the distance from a point perpendicular to path,
		 * has the path in some cases be handled in several sequence,
		 * this returns an array of values where to do this splitting */
		protected function turnpoints():Array {
			var fx:Array = this.bezierEquation("x"),
				fy:Array = this.bezierEquation("y"),
				// f(x)= 	A +Bx +Cxx +Dxxx
				// f'(x)= 	B +2Cx +3Dxx
				// f''(x)=	2C +6Dx = 0
				x0:Number = -fx[2] / (3*fx[3]),
				y0:Number = -fy[2] / (3*fy[3]),
				arr:Array = [];
			if (0<x0 && x0<1) arr.push(x0);
			if (0<y0 && y0<1) {
				arr.push(y0);
				if (arr.length==2) {
					arr = x0<y0 ? [x0,y0] : [y0,x0];
				}
			}
			
			return arr;
		}
		
		
		/**	Set precision,
		 * if 'true' renders the path in several sequences, split by the zeroes of the second derivative */
		public function set divided(value:Boolean):void {
			this._divided=value; 
			update();
		}
		[Bindable]
		public function get divided():Boolean { 
			return this._divided;
		}
		
		
		/**	true: repeats the image along the path
		 *  false: image rendered only at one location
		 **/
		public function set repeat(value:Boolean):void {
			shader.data.imagewidth.value= [ value ? this._input.width : 0.0 ]; 
			update();
		}
		[Bindable]
		public function get repeat():Boolean { 
			var t:Number = shader.data.imagewidth.value[0];
			return t>0.5;
		}
		
		
		/**	Sets texture scaling */
		public function set scale(value:Number):void {
			shader.data.scale.value= [value, value];
			this._fit=0;
			update();
		}
		[Bindable]
		public function get scale():Number { 
			return shader.data.scale.value[0];
		}
		
		
		/**	Controls how many time the image is repeated on the path */
		public function set fit(value:uint):void {
			this._fit = value;
			if (_fit>0) shader.data.offset.value[0] = 0.0;
			update();
		}
		[Bindable]
		public function get fit():uint { 
			return _fit;
		}
		
		
		/**	Sets the pattern rotation angle, value in degrees */
		public function set rotation(degrees:Number):void { 
			var radians:Number = degrees*Math.PI/180,
				cs:Number = Math.cos(radians),
				sn:Number = Math.sin(radians);
			//	set the rotation matrix
			shader.data.rotation.value  = [ cs,sn,-sn,cs ];
			
			update();
		}
		
		
		public function set offset(point:Object):void { 
			shader.data.offset.value = [ point.x, point.y ]; 
			update();
		}
		public function set offsetX(value:Number):void {
			shader.data.offset.value[0] = value; 
			update();
		}
		[Bindable]
		public function get offsetX():Number {
			return shader.data.offset.value[0];
		}
		
		
		public function set offsetY(value:Number):void {
			shader.data.offset.value[1] = value; 
			update();
		}
		[Bindable]
		public function get offsetY():Number {
			return shader.data.offset.value[1];
		}
		
		
	}
}